"""This software is provide 'as is' and the National Park Service gives no
warranty, expressed or implied, as to the accuracy, reliability, or
completeness of this software. Although this software has been processed
successfully on a computer system at the National Park Service, no warranty
expressed or implied is made regarding the functioning of the software on
another system or for general or scientific purposes, nor shall the act of
distribution constitute any such warranty. This disclaimer applies both to
individual use of the software and aggregate use with other software.
"""
import sys
import os
import arcpy 
import numpy 
import datetime
import math
from arcpy.sa import *

class GammaReadings(object):
    def __init__(self):
        self.surveyLineData = {}

    def getLineNumbers(self, data):
        """Returns a set of line numbers from the data dictionary
        Arguments: 
        data - a list of dictionaries
        """
        lineNumberList = []
        for record in data:
            lineNumberList.append(record['lineNumber'])
        lineNumberSet = set(lineNumberList)
        return lineNumberSet

    def createSurveyLineData(self, data):
        """Splits the data dictionary up by line number and populates
        the surveyLineData dictionary.
        Arguments: 
        data - a list of dictionaries
        """
        lineNumbers = self.getLineNumbers(data)
        self.surveyLineData = {}
        for lineNumber in lineNumbers:
            self.surveyLineData['{}'.format(lineNumber)] = [record for record in data if record['lineNumber'] == lineNumber]		

    def removeOutlyingValuesSD(self, key):
        """Removes records that have a value outside two standard deviations of
        the mean for any key in the surveyLineData dictionary.  All records in all survey
        lines are used in the analysis.
        Arguments:
        data - list of dictionaries - list of dictionaries
        key - sting - key in the data dictionary to filter
        """
        for surveyLine in self.surveyLineData:
            numpyValues = numpy.array([float(record[key]) for record in self.surveyLineData[surveyLine]])
            standardDeviation = numpy.std(numpyValues)
            mean = numpy.mean(numpyValues)
            maxValue = mean + ( 2.0 * standardDeviation)
            minValue = mean - ( 2.0 * standardDeviation)
            self.surveyLineData[surveyLine] = [record for record in self.surveyLineData[surveyLine] if (float(record[key]) >= minValue and float(record[key]) <= maxValue)]

    def removeValueChangesGreaterThanNeighborhoodMean(self, key, change, back, forward):
        """Removes records that have a change in a key value greater than
        the specified allowable change from the mean of a neighborhood of points.
        Arguments:
        key - string - key in the surveyLineData dictionary to filter
        change - float - the maximum allowable change from record to record
        back - int - number previous records to include in the neighborhood mean value
        forward - int - number of following records to include in the negighborhood mean value
        """        
        for surveyLine in self.surveyLineData:
            for recordIndex in range(len(self.surveyLineData[surveyLine])):
                values = []
                neighborhood = range(recordIndex - back, recordIndex + forward + 1)
                neighborhood = [index for index in neighborhood if index != recordIndex]
                for searchIndex in neighborhood:
                    if searchIndex >= 0 and searchIndex < len(self.surveyLineData[surveyLine]):
                        values.append(self.surveyLineData[surveyLine][searchIndex][key])
                
                if len(values) > 0:
                    numpyValues = numpy.array([float(record) for record in values])
                    meanValue = numpyValues.mean()
                    deltaValue = self.surveyLineData[surveyLine][recordIndex][key] - meanValue
                    if abs(deltaValue) > change:
                        self.surveyLineData[surveyLine][recordIndex]['filter'] = True
                    else:
                        self.surveyLineData[surveyLine][recordIndex]['filter'] = False
                else:
                    self.surveyLineData[surveyLine][recordIndex]['filter'] = False                
            self.surveyLineData[surveyLine] = [record for record in self.surveyLineData[surveyLine] if record['filter'] == False]

    def calculateDeltaGamma(self, sensorNoise, back, forward):
        """Calculates the change in gamma as the difference between the current and
        the mean of the surrounding gamma readings. If the change in
        gamma is less than the sensor noise it is reassigned the value of 0 and the
        change in gamma is assumed to be the result of noise. Does not maintain neighborhood
        size at the survey line ends.
        Aruguments: 
        sensorNoise - float - sensor noise in gamma
        back - int - number previous records to include in the mean background gamma
        forward - int - number of following records to include in the mean background gamma
        """		
        for surveyLine in self.surveyLineData:
            for recordIndex in range(len(self.surveyLineData[surveyLine])):
                values = []
                neighborhood = range(recordIndex - back, recordIndex + forward + 1)
                neighborhood = [index for index in neighborhood if index != recordIndex]
                for searchIndex in neighborhood:
                    if searchIndex >= 0 and searchIndex < len(self.surveyLineData[surveyLine]):
                        values.append(self.surveyLineData[surveyLine][searchIndex]['recordedGamma'])
                if len(values) > 0:
                    numpyValues = numpy.array([float(record) for record in values])
                    meanValue = numpyValues.mean()
                    deltaGamma = self.surveyLineData[surveyLine][recordIndex]['recordedGamma'] - meanValue
                    if abs(deltaGamma) < sensorNoise:
                        deltaGamma = 0
                    self.surveyLineData[surveyLine][recordIndex]['deltaGamma'] = deltaGamma
                else:
                    self.surveyLineData[surveyLine][recordIndex]['deltaGamma'] = 0

    def calculateSurveyLineGradients(self):
        
        i = 0
        for surveyLine in self.surveyLineData:
            i += len(self.surveyLineData[surveyLine]) - 1

        numpyGradientData = numpy.zeros(i, dtype={'names':['easting', 'northing', 'gradientGamma', 'lineNumber'], 'formats':['f8', 'f8', 'f8', 'a32']})

        index = 0
        for surveyLine in self.surveyLineData:
            multiplier = 0.0
            survey_line_len = len(self.surveyLineData[surveyLine])
            index_start = index
            index_stop = 0
            for recordIndex in range(1, survey_line_len):
                item2 = self.surveyLineData[surveyLine][recordIndex]
                item1 = self.surveyLineData[surveyLine][recordIndex - 1]
                difference = item2['recordedGamma'] - item1['recordedGamma']
                distance = math.sqrt(math.pow(item2['easting'] - item1['easting'], 2) + math.pow(item2['northing'] - item1['northing'], 2))

                numpyGradientData[index]['easting'] = (item2['easting'] + item1['easting']) / 2.0
                numpyGradientData[index]['northing'] = (item2['northing'] + item1['northing']) / 2.0
                numpyGradientData[index]['gradientGamma'] = difference / distance
                numpyGradientData[index]['lineNumber'] = item2['lineNumber']

                index += 1
                index_stop = index
                multiplier += distance

            multiplier = multiplier / (survey_line_len - 1)

            #arcpy.AddMessage('surveyLine: ' + str(surveyLine) + ', length: ' + str(survey_line_len) + ', multiplier: ' + str(multiplier))
            for recordIndex in range(index_start, survey_line_len - 1):
                numpyGradientData[recordIndex]['gradientGamma'] = numpyGradientData[recordIndex]['gradientGamma'] * multiplier
                
        return numpyGradientData

    def createNumpyArray(self, includeDeltaGamma):
        """Creates a numpy structured array from the surveyLineData dictionary
        Arguments: 
        includeDeltaGamma - bool - Set to true to include deltaGamma key
        Returns: 
        numpy structured array
        """
        data=[]				
        for surveyLine in self.surveyLineData:
            for record in self.surveyLineData[surveyLine]:
                data.append(record)
        if includeDeltaGamma:
            numpyData = numpy.zeros(len(data), dtype={'names':['easting', 'northing', 'recordedGamma', 'lineNumber', 'altitude', 'deltaGamma'], 'formats':['f8', 'f8', 'f8', 'a32', 'i4', 'i4']})
            for index in range(len(data)):
                numpyData[index]['easting'] = data[index]['easting']
                numpyData[index]['northing'] = data[index]['northing']
                numpyData[index]['recordedGamma'] = data[index]['recordedGamma']
                numpyData[index]['lineNumber'] = data[index]['lineNumber']
                numpyData[index]['altitude'] = int(round(data[index]['altitude']))
                numpyData[index]['deltaGamma'] = int(round(data[index]['deltaGamma']))

        else:
            numpyData = numpy.zeros(len(data), dtype={'names':['easting', 'northing', 'recordedGamma', 'lineNumber', 'altitude'], 'formats':['f8', 'f8', 'f8', 'a32', 'i4']})
            for index in range(len(data)):
                numpyData[index]['easting'] = data[index]['easting']
                numpyData[index]['northing'] = data[index]['northing']
                numpyData[index]['recordedGamma'] = data[index]['recordedGamma']
                numpyData[index]['lineNumber'] = data[index]['lineNumber']
                numpyData[index]['altitude'] = int(round(data[index]['altitude']))			

        return numpyData


    def readDataTableWithAltitude(self, inputTable, eastingField,
                                  northingField, recordedGammaField, lineNumberField, altitudeField,
                                  altitudeUnit):
        """Reads a table with an arcpy search cursor and populates a dictionary
        for recorded magnetometer datasets using altitude.
        Arguments: 
        inputTable - string - path to a table
        eastingField - string - easting field name
        northingField - string - northing field name
        recordedGammaField - string - recorded gamma field name
        lineNumberField - string - line number field name
        altitudeField - string - altitude field name
        altitudeUnit - string - altitude linear unit, either meter or feet
        """
        data = []
        unitConversions = {'meter':100, 'feet':30.48}
        with arcpy.da.SearchCursor(inputTable, [eastingField,
                                                northingField,
                                                recordedGammaField,
                                                lineNumberField,
                                                altitudeField]) as cursor:
            for row in cursor:
                data.append({'easting':row[0],
                             'northing':row[1],
                             'recordedGamma':row[2],
                             'lineNumber':row[3],
                             'altitude':row[4] * unitConversions[altitudeUnit]})
        self.createSurveyLineData(data)

    def readDataTableWithOutAltitude(self, inputTable, eastingField,
                                     northingField, recordedGammaField, lineNumberField, sensorAltitude):
        """Reads a table with an arcpy search cursor and populates a dictionary
        for recorded magnetometer datasets without using altitude.
        Arguments: 
        inputTable - string - path to a table
        eastingField - string - easting field name
        northingField - string - northing field name
        recordedGammaField - string - recorded gamma field name
        lineNumberField - string - line number field name
        sensorAltitude - double - constant sensor altitude in centimeters
        """
        data = []
        unitConversions = {'meter':100, 'feet':30.48}
        with arcpy.da.SearchCursor(inputTable, [eastingField,
                                                northingField,
                                                recordedGammaField,
                                                lineNumberField]) as cursor:
            for row in cursor:
                data.append({'easting':row[0],
                             'northing':row[1],
                             'recordedGamma':row[2],
                             'lineNumber':row[3],
                             'altitude':sensorAltitude})
        self.createSurveyLineData(data)

    def readSurveyLinesFromFeatureClass(self, inputFeatureClass):

        data = []
        with arcpy.da.SearchCursor(inputFeatureClass, ['SHAPE@X', 'SHAPE@Y', 'recordedGamma', 'lineNumber', 'altitude']) as cursor:
            for row in cursor:
                data.append({'easting':row[0], 
                             'northing':row[1], 
                             'recordedGamma':row[2],
                             'lineNumber':row[3],
                             'altitude':row[4]})
        self.createSurveyLineData(data)

class Toolbox(object): 
    def __init__(self): 
        """Define the toolbox (the name of the toolbox is the name of the .pyt file)."""
        self.label = 'Magnetometer Survey Toolbox'
        self.alias = '' 

        # List of tool classes associated with this toolbox 
        self.tools = [InputTool, ConfidenceTool, VisualizationTool, SurveyAreaTool] 

class InputTool(object):
    def __init__(self):
        self.label = 'Input Tool'    
        self.description = ''
        self.canRunInBackground = True

    def getParameterInfo(self):

        param0 = arcpy.Parameter(displayName = 'Data Table', 
                                 name = 'inputTable', 
                                 datatype = 'Table', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param1 = arcpy.Parameter(displayName = 'Easting Field', 
                                 name = 'inputEasting', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param2 = arcpy.Parameter(displayName = 'Northing Field', 
                                 name = 'inputNorthing', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Gamma Field', 
                                 name = 'inputGamma', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param4 =  arcpy.Parameter(displayName = 'Altitude values included in table?', 
                                  name = 'useAltitudeField', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 
        
        param5 = arcpy.Parameter(displayName = 'Altitude Field', 
                                 name = 'inputAltitudeField', 
                                 datatype = 'Field', 
                                 parameterType = 'Optional', 
                                 direction = 'Input')

        param6 = arcpy.Parameter(displayName = 'Altitude is in Feet (otherwise in meters)', 
                                 name = 'useFeetAsAltitude', 
                                 datatype = 'GPBoolean', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Enter estimated altitude of sensor', 
                                  name = 'inputSensorAltitude', 
                                  datatype = 'Double', 
                                  parameterType = 'Optional', 
                                  direction = 'Input') 

        param8 = arcpy.Parameter(displayName= 'Line Number Field', 
                                 name = 'inputLineNumber', 
                                 datatype = 'Field', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param9 = arcpy.Parameter(displayName = 'Spatial Reference', 
                                 name = 'inputSpatialReference', 
                                 datatype = 'Spatial Reference', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param10 = arcpy.Parameter(displayName = 'New Output File Geodatabase',
                                  name = 'inputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Output') 

        # Default values 
        param1.value = "Field1"
        param2.value = "Field2"
        param3.value = "Field3"
        param4.value = True
        param5.value = "Field4"
        param6.value = True
        param7.value = 3
        param8.value = "Field8"
        param9.value = "NAD 1983 UTM Zone 17N"

        param1.parameterDependencies = [param0.name] 
        param2.parameterDependencies = [param0.name] 
        param3.parameterDependencies = [param0.name] 
        param5.parameterDependencies = [param0.name] 
        param8.parameterDependencies = [param0.name]
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, param9, param10] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters):
        if parameters[4].value == True:
            parameters[5].enabled = True
            parameters[7].enabled = False

        if parameters[4].value == False:
            parameters[5].enabled = False
            parameters[7].enabled = True

        return

    def updateMessages(self, parameters):
        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input parameters 
        inputTable = parameters[0].valueAsText 
        arcpy.AddMessage('Table: {}'.format(inputTable))

        inputEastingField = parameters[1].valueAsText 
        arcpy.AddMessage('Easting: {}'.format(inputEastingField))

        inputNorthingField = parameters[2].valueAsText 
        arcpy.AddMessage('Northing: {}'.format(inputNorthingField)) 

        inputGammaField = parameters[3].valueAsText 
        arcpy.AddMessage('Gamma: {}'.format(inputGammaField)) 

        useAltitudeField = parameters[4].value 
        arcpy.AddMessage('Use altitude field? {}'.format(str(useAltitudeField))) 
        
        if useAltitudeField:
            inputAltitudeField = parameters[5].valueAsText 
            arcpy.AddMessage('Altitude Field: {}'.format(inputAltitudeField))
        
        useFeetAsAltitude = parameters[6].value 
        arcpy.AddMessage('Altitude is measured in feet? {}'.format(str(useFeetAsAltitude))) 
        
        if useAltitudeField == False:
            unitConversions = {'meter':100, 'feet':30.48}
            if useFeetAsAltitude:
                inputSensorAltitude = parameters[7].value * unitConversions['feet']
                arcpy.AddMessage('Input sensor altitude: {}'.format(str(inputSensorAltitude)))
            else:
                inputSensorAltitude = parameters[7].value * unitConversions['meter']
                arcpy.AddMessage('Input sensor altitude: {}'.format(str(inputSensorAltitude)))

        inputLineNumberField = parameters[8].valueAsText 
        arcpy.AddMessage('Line Number Field: {}'.format(inputLineNumberField)) 

        inputSpatialReference = parameters[9].value 
        arcpy.AddMessage('Spatial reference: {}'.format(inputSpatialReference.name)) 

        outputWorkspace = parameters[10].valueAsText

        if not outputWorkspace.endswith('.gdb'):
            outputWorkspace = outputWorkspace + '.gdb'

        workspaceparts = os.path.split(outputWorkspace)

        arcpy.AddMessage('Adding Geodatabase: ' + workspaceparts[0] + ': ' + workspaceparts[1]) 

        arcpy.CreateFileGDB_management (workspaceparts[0], workspaceparts[1])

        # Create gammaReadings object 
        gammaReadings = GammaReadings() 

        # Read the data from the table 
        if useAltitudeField: 
            if useFeetAsAltitude: 
                gammaReadings.readDataTableWithAltitude(inputTable, inputEastingField,
                                                        inputNorthingField, inputGammaField,
                                                        inputLineNumberField, inputAltitudeField, 'feet')
            else: 
                gammaReadings.readDataTableWithAltitude(inputTable, inputEastingField,
                                                        inputNorthingField, inputGammaField,
                                                        inputLineNumberField, inputAltitudeField, 'meter') 
        else:
            if useFeetAsAltitude:
                gammaReadings.readDataTableWithOutAltitude(inputTable,
                                                           inputEastingField, inputNorthingField, inputGammaField,
                                                           inputLineNumberField, inputSensorAltitude)
            else:
                gammaReadings.readDataTableWithOutAltitude(inputTable,
                                                           inputEastingField, inputNorthingField, inputGammaField,
                                                           inputLineNumberField, inputSensorAltitude)
        arcpy.AddMessage('Read the data table')

        # Create a feature class from the gamma readings 
        gammaNumpyArray = gammaReadings.createNumpyArray(False) 

        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True

        outputFeatureClass = os.path.join(outputWorkspace, 'rawGamma')
        arcpy.da.NumPyArrayToFeatureClass(gammaNumpyArray, outputFeatureClass, ['easting', 'northing'], inputSpatialReference) 
        arcpy.AddMessage('Created rawGamma') 
        arcpy.PointsToLine_management('rawGamma', 'surveyLines', 'lineNumber')
        arcpy.AddMessage('Created surveyLines') 


class ConfidenceTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class."""
        self.label = 'Confidence Model'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""

        param0 = arcpy.Parameter(displayName = 'Input Survey Points Feature Class', 
                                 name = 'inputSurveyPoints', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 
        
        param1 = arcpy.Parameter(displayName = 'Filter elevations by standard deviations? (removes elevations outside of 2 standard deviations from the mean)', 
                                  name = 'filterElevationsSD', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param2 =  arcpy.Parameter(displayName = 'Filter elevations by maximum change from the neighborhood mean?', 
                                  name = 'filterElevationsMaxChange', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Points back to include in neighborhood for maximum change elevation filter ', 
                                  name = 'inputElevationPointsBack', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param4 = arcpy.Parameter(displayName = 'Points forward to include in neighborhood for maximum change elevation filter', 
                                  name = 'inputElevationPointsForward', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param5 = arcpy.Parameter(displayName = 'Maximum elevation change from neighborhood average (centimeters)', 
                                 name = 'maximumChange', 
                                 datatype = 'Double', 
                                 parameterType = 'Optional', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Survey Area', 
                                  name = 'inputSurveyArea', 
                                  datatype = 'Feature Class', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param7 = arcpy.Parameter(displayName = 'Magnetic Moment', 
                                  name = 'inputMagneticMoment', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param8 = arcpy.Parameter(displayName = 'Sensor Noise', 
                                  name = 'inputSensorNoise', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param9 = arcpy.Parameter(displayName = 'Mass Threshold', 
                                  name = 'inputMassThreshold', 
                                  datatype = 'Double', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param10 = arcpy.Parameter(displayName = 'Points back to include in neighborhood for delta gamma calculation', 
                                  name = 'inputDeltaGammaPointsBack', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param11 = arcpy.Parameter(displayName = 'Points forward to include in neighborhood for delta gamma calculation', 
                                  name = 'inputDeltaGammaPointsForward', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param12 = arcpy.Parameter(displayName = 'Cell size for output rasters', 
                                  name = 'inputCellSize', 
                                  datatype = 'GPLong', 
                                  parameterType = 'Required', 
                                  direction = 'Input')

        param13 =  arcpy.Parameter(displayName = 'Send Outputs to same geodatabase as input?', 
                                  name = 'useInputGeodatabase', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param14 = arcpy.Parameter(displayName = 'Existing Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param1.value = False
        param2.value = False
        param3.value = 1
        param4.value = 1
        param5.value = 1
        param7.value = 50
        param8.value = 3
        param9.value = 50
        param10.value = 1
        param11.value = 1
        param12.value = 1
        param13.value = True

        # Dependencies 
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8, 
                param9, param10, param11, param12, param13, param14] 


    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.CheckExtension('Spatial') != 'Available': 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        """Modify the values and properties of parameters before internal 
        validation is performed.  This method is called whenever a parameter 
        has been changed."""
        if parameters[2].value == True:
            parameters[3].enabled = True
            parameters[4].enabled = True
            parameters[5].enabled = True
            
        if parameters[2].value == False:
            parameters[3].enabled = False
            parameters[4].enabled = False
            parameters[5].enabled = False

        if (parameters[13].value == True):
            parameters[14].enabled = False

            if (parameters[0].altered and len(parameters[0].valueAsText) > 0) : 
                workspaceparts = os.path.split(parameters[0].valueAsText)
                parameters[14].value = workspaceparts[0]
            else :
                parameters[14].value = ''
        else :
            parameters[14].enabled = True

        return

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                workspaceparts = os.path.split(inputFeatureClass)
            
                arcpy.env.workspace = workspaceparts[0]
                desc = arcpy.Describe(inputFeatureClass)

                if (desc.shapeType != 'Point'):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be points feature class')
                else :
                    fieldList = arcpy.ListFields(inputFeatureClass)
                    b_gamma = False
                    b_line = False

                    for field in fieldList:
                        if field.name == 'recordedGamma':
                            b_gamma = True
                        elif field.name == 'lineNumber':
                            b_line = True

                    if b_gamma == False or b_line == False or b_alt == False:
                        parameters[0].setErrorMessage('Not all required input fields present lineNumber and recordedGamma')

        except Exception as e:
            return
        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input parameters 
        inputFeatureClass = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputFeatureClass))

        filterElevationsSD = parameters[1].value
        arcpy.AddMessage('Filter elevations based on Standard Deviations? {}'.format(str(filterElevationsSD)))

        filterElevationsMaxChange = parameters[2].value 
        arcpy.AddMessage('Filter elevations based on maximum change? {}'.format(str(filterElevationsMaxChange))) 

        if filterElevationsMaxChange:
            inputElevationPointsBack = int(parameters[3].value)
            arcpy.AddMessage('Elevation points back: {}'.format(str(inputElevationPointsBack)))
            
            inputElevationPointsForward = int(parameters[4].value)
            arcpy.AddMessage('Elevation points forward: {}'.format(str(inputElevationPointsForward)))
            
            maximumChange = parameters[5].value
            arcpy.AddMessage('Maximum elevation change from neighborhood mean: {}'.format(str(maximumChange)))
    
        inputSurveyArea = parameters[6].valueAsText 
        arcpy.AddMessage('Survey area: {}'.format(inputSurveyArea)) 

        inputMagneticMoment = parameters[7].value
        arcpy.AddMessage('Magnetic moment: {}'.format(str(inputMagneticMoment))) 

        inputSensorNoise = parameters[8].value
        arcpy.AddMessage('Sensor noise: {}'.format(str(inputSensorNoise))) 

        inputMassThreshold = parameters[9].value
        arcpy.AddMessage('Mass threshold: {}'.format(str(inputMassThreshold)))

        inputDeltaGammaPointsBack = int(parameters[10].value)
        arcpy.AddMessage('Points back for delta gamma calculation: {}'.format(str(inputDeltaGammaPointsBack)))

        inputDeltaGammaPointsForward = int(parameters[11].value)
        arcpy.AddMessage('Points forward for delta gamma calculation: {}'.format(str(inputDeltaGammaPointsForward)))

        inputCellSize = parameters[12].value
        arcpy.AddMessage('Cell size for output rasters: {}'.format(str(inputCellSize)))

        outputWorkspace = parameters[14].valueAsText

        arcpy.AddMessage('outputWorkspace: {}'.format(outputWorkspace))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.env.mask = inputSurveyArea 
        arcpy.env.extent = inputSurveyArea

        arcpy.AddMessage('Set environment variables')         

        suffix = ''
        delta_suffix = ''
        delta_suffix_2 = ''

        if filterElevationsSD:
            delta_suffix = delta_suffix + '_FSD'

        if filterElevationsMaxChange:
            delta_suffix = delta_suffix + '_FMC_' + str(inputElevationPointsBack) + '_' + str(inputElevationPointsForward) + '_' + str(int(maximumChange))

        delta_suffix = delta_suffix + '_N_' + str(int(inputSensorNoise)) + '_DR_' + str(int(inputDeltaGammaPointsBack)) + '_' + str(int(inputDeltaGammaPointsForward)) 
        delta_suffix_2 = delta_suffix + '_CS_' + str(int(inputCellSize)) 

        suffix = delta_suffix_2 + '_MM_' + str(int(inputMagneticMoment)) + '_T_' + str(int(inputMassThreshold)) 
        
        # Create the logTable
        now = datetime.datetime.now()
        logTable = 'ConfidenceLog' + suffix
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.AddField_management(logTable, 'Process', 'TEXT')
        arcpy.AddField_management(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
        if filterElevationsSD:
            logCursor.insertRow(('Filter elevations based on Standard Deviations? {}'.format(str(filterElevationsSD)), datetime.datetime.now()))

        if filterElevationsMaxChange:
            logCursor.insertRow(('Filter elevations based on maximum change? {}'.format(str(filterElevationsMaxChange)), datetime.datetime.now()))
            logCursor.insertRow(('Elevation points back: {}'.format(str(inputElevationPointsBack)), datetime.datetime.now()))
            logCursor.insertRow(('Elevation points forward: {}'.format(str(inputElevationPointsForward)), datetime.datetime.now()))
            logCursor.insertRow(('Maximum elevation change from neighborhood mean: {}'.format(str(maximumChange)), datetime.datetime.now()))
            
        logCursor.insertRow(('Survey area: {}'.format(inputSurveyArea), datetime.datetime.now()))
        logCursor.insertRow(('Magnetic moment: {}'.format(str(inputMagneticMoment)), datetime.datetime.now()))
        logCursor.insertRow(('Sensor noise: {}'.format(str(inputSensorNoise)), datetime.datetime.now()))
        logCursor.insertRow(('Mass threshold: {}'.format(str(inputMassThreshold)), datetime.datetime.now()))
        logCursor.insertRow(('Points back for delta gamma calculation: {}'.format(str(inputDeltaGammaPointsBack)), datetime.datetime.now()))
        logCursor.insertRow(('Points forward for delta gamma calculation: {}'.format(str(inputDeltaGammaPointsForward)), datetime.datetime.now()))
        logCursor.insertRow(('Cell size for output rasters: {}'.format(str(inputCellSize)), datetime.datetime.now()))

        # Check out the spatial analyst extension 
        arcpy.CheckOutExtension('Spatial') 
        arcpy.AddMessage('Checked out Spatial Analyst extension') 
        logCursor.insertRow(('Checked out Spatial Analyst extension', datetime.datetime.now()))

        # Create gammaReadings object 
        gammaReadings = GammaReadings() 

        # Read the data from the table 
        gammaReadings.readSurveyLinesFromFeatureClass(inputFeatureClass)

        arcpy.AddMessage('Read the data table')
        logCursor.insertRow(('Read the data table', datetime.datetime.now()))

        # Filter elevation values outside of 2 standard deviations
        if filterElevationsSD:
            gammaReadings.removeOutlyingValuesSD('altitude')
            arcpy.AddMessage('Filtered altitude values based on standard deviations')
            logCursor.insertRow(('Filtered altitude values based on standard deviations', datetime.datetime.now()))

        # Filter elevation values greater than the neighborhood mean if selected 
        if filterElevationsMaxChange:
            gammaReadings.removeValueChangesGreaterThanNeighborhoodMean('altitude',
                                                                        maximumChange, inputElevationPointsBack,
                                                                        inputElevationPointsForward)
            arcpy.AddMessage('Filtered elevations based on maximum change') 
            logCursor.insertRow(('Filtered elevations based on maximum change', datetime.datetime.now()))

        # Calculate delta gamma values 
        gammaReadings.calculateDeltaGamma(inputSensorNoise, inputDeltaGammaPointsBack, inputDeltaGammaPointsForward)
        arcpy.AddMessage('Calculated deltaGamma') 
        logCursor.insertRow(('Calculated deltaGamma', datetime.datetime.now()))

        inputSpatialReference = arcpy.Describe(inputFeatureClass).spatialReference

        # Create a feature class from the gamma readings 
        gammaNumpyArray = gammaReadings.createNumpyArray(True)
        deltaGamma =  'deltaGamma' + delta_suffix
        outputFeatureClass = os.path.join(outputWorkspace, deltaGamma)
        if arcpy.Exists(outputFeatureClass):
            arcpy.Delete_management(outputFeatureClass)
        arcpy.da.NumPyArrayToFeatureClass(gammaNumpyArray, outputFeatureClass, ['easting', 'northing'], inputSpatialReference) 
        arcpy.AddMessage('Created deltaGamma') 
        logCursor.insertRow(('Created deltaGamma', datetime.datetime.now()))

        del gammaNumpyArray
        del gammaReadings

        # Allocate deltaGamma Values as a raster 
        arcpy.AddMessage('Set Output Extent')
        deltaGammaSurface = EucAllocation(deltaGamma, '', '', inputCellSize, 'deltaGamma') 
        deltaGammaSurface.save('deltaGamma' + delta_suffix_2) 
        arcpy.AddMessage('Created Delta Gamma Surface')
        logCursor.insertRow(('Created Delta Gamma Surface', datetime.datetime.now()))

        # Create the absolute value of deltaGamma
        absDeltaGammaSurface = Abs(deltaGammaSurface) 
        absDeltaGammaSurface.save('absDeltaGamma' + delta_suffix_2) 
        arcpy.AddMessage('Created abs Delta Gamma Surface') 
        logCursor.insertRow(('Created abs Delta Gamma Surface', datetime.datetime.now()))

        # Set the geoprocessing snap raster to absDeltaGamma 
        arcpy.env.snapRaster = 'absDeltaGamma' + delta_suffix_2
        arcpy.AddMessage('Set snap raster') 
        logCursor.insertRow(('Set snap raster', datetime.datetime.now()))

        # Allocate altitude values as a raster 
        altitudeCentimeters = EucAllocation(deltaGamma, '', '', inputCellSize, 'altitude') 
        altitudeCentimeters.save('altitudeCentimeters' + delta_suffix_2) 
        arcpy.AddMessage('Created altitude') 
        logCursor.insertRow(('Created altitude', datetime.datetime.now()))

        # Create a horizontal euclidean distance raster from the input source points 
        horizontalDistance = EucDistance(deltaGamma, '', inputCellSize) 

        horizontalCentimeters = Int(horizontalDistance * inputSpatialReference.metersPerUnit * 100) 
        horizontalCentimeters.save('horizontalCentimeters' + delta_suffix_2) 
        arcpy.AddMessage('Created horizontalCentimeters') 
        logCursor.insertRow(('Created horizontalCentimeters', datetime.datetime.now()))

        # Calculate the final distance raster in centimeters 
        centimeters = Int((horizontalCentimeters**2 + altitudeCentimeters**2)**0.5) 
        centimeters.save('centimeters' + delta_suffix_2) 
        arcpy.AddMessage('Created centimeters with altitude') 
        logCursor.insertRow(('Created centimeters with altitude field values', datetime.datetime.now()))

        # Create mass (kg) of potentially missed objects 
        potentialMassMissed = Int((inputSensorNoise * 0.000001 * (centimeters)**3) / (inputMagneticMoment)) 
        potentialMassMissed.save('potentialMassMissed' + suffix) 
        arcpy.AddMessage('Created potentialMassMissed') 
        logCursor.insertRow(('Created potentialMassMissed', datetime.datetime.now()))

        try:
            arcpy.BuildRasterAttributeTable_management('potentialMassMissed' + suffix, "Overwrite")
        except:
            arcpy.AddMessage("Build Raster Attribute Table failed for potential mass missed.")

        # Create mass (kg) of potentially found objects 
        potentialMassObserved = Int((absDeltaGammaSurface * 0.000001 * (centimeters)**3) / (inputMagneticMoment)) 
        potentialMassObserved.save('potentialMassObserved' + suffix) 
        arcpy.AddMessage('Created potentialMassObserved') 
        logCursor.insertRow(('Created potentialMassObserved', datetime.datetime.now()))

        try:
            arcpy.BuildRasterAttributeTable_management('potentialMassObserved' + suffix, "Overwrite")
        except:
            arcpy.AddMessage("Build Raster Attribute Table failed for potential mass observed.")


        # Reclassify potential mass missed raster
        if inputMassThreshold > potentialMassMissed.minimum:
            remapRange = RemapRange([[0,inputMassThreshold, 1], [inputMassThreshold, 9999999, 0]])
            confidenceResult = Reclassify(potentialMassMissed, 'Value', remapRange) 
            confidenceResult.save('confidenceResult' + suffix)
        else:
            remapRange = RemapRange([[inputMassThreshold, 9999999, 0]])
            confidenceResult = Reclassify(potentialMassMissed, 'Value', remapRange) 
            confidenceResult.save('confidenceResult' + suffix)           
        arcpy.AddMessage('Created Confidence Result')
        logCursor.insertRow(('Created Confidence Result', datetime.datetime.now()))

        # Display the percentage of the survey area adequatly surveyed
        fields = ['VALUE', 'COUNT']
        data = {r[0]:r[1] for r in arcpy.da.SearchCursor('confidenceResult' + suffix, fields)}
        keys = data.keys()
        if 0 in keys:
            if 1 in keys:
                totalCount = data[0] + data[1]
                percentCovered = (data[1] / totalCount)
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequatly Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequatly Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

            else:
                percentCovered = 0
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequatly Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequatly Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

        elif 0 not in keys:
            if 1 in keys:
                percentCovered = 1
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('Percent Adequatly Surveyed: \n{:.2%}'.format(percentCovered))
                arcpy.AddMessage('*****************')
                arcpy.AddMessage('*****************')
                logCursor.insertRow(('Percent Adequatly Surveyed: {:.2%}'.format(percentCovered), datetime.datetime.now()))

        # Delete the logCursor object
        del logCursor
        arcpy.CheckInExtension('Spatial') 

class VisualizationTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)"""
        self.label = 'Visualization'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Survey Points Feature Class', 
                                 name = 'inputSurveyPoints', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param1 = arcpy.Parameter(displayName = 'Select a Surface creation Method',
                                 name = 'surfaceMethod',
                                 datatype = 'GPString',
                                 parameterType = 'Required',
                                 direction = 'Input')

        param2 = arcpy.Parameter(displayName = 'Survey Area', 
                                 name = 'inputSurveyArea', 
                                 datatype = 'Feature Class', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Survey Line Spacing (meter)', 
                                 name = 'inputLineSpacing', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param4 = arcpy.Parameter(displayName = 'Raw Gamma Contour Interval (gamma)', 
                                 name = 'inputRawContourInterval', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param5 = arcpy.Parameter(displayName = 'Gradient Gamma Contour Interval (gamma)', 
                                 name = 'inputGradientContourInterval', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param6 = arcpy.Parameter(displayName = 'Cell Size for output rasters',
                                 name = 'inputCellSize',
                                 datatype = 'GPLong',
                                 parameterType = 'Required',
                                 direction = 'Input')

        param7 =  arcpy.Parameter(displayName = 'Send Outputs to same geodatabase as input?', 
                                  name = 'useInputGeodatabase', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param8 = arcpy.Parameter(displayName = 'Existing Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param3.value = 10.0
        param4.value = 2.0
        param5.value = 1.0
        param6.value = 1
        param7.value = True

        param1.filter.type = "ValueList"

        if arcpy.CheckExtension('3D') != "Available":
            param1.filter.list = ["Kriging"]
            param1.value = "Kriging"
        else:
            param1.filter.list = ["Kriging", "TIN"]
            param1.value = "TIN"

        # Dependencies 
        return [param0, param1, param2, param3, param4, param5, param6, param7, param8] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.CheckExtension('Spatial') != "Available": 
                raise Exception
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        """Modify the values and properties of parameters before internal 
        validation is performed.  This method is called whenever a parameter 
        has been changed."""

        if (parameters[1].value == "Kriging") :
            parameters[3].enabled = True
        else :
            parameters[3].enabled = False

        if (parameters[4].value < 0) :
            parameters[4].value = abs(parameters[4].value)

        if (parameters[5].value < 0) :
            parameters[5].value = abs(parameters[5].value)

        if (parameters[7].value == True):
            parameters[8].enabled = False

            if (parameters[0].altered and len(parameters[0].valueAsText) > 0) : 
                workspaceparts = os.path.split(parameters[0].valueAsText)
                parameters[8].value = workspaceparts[0]
            else :
                parameters[8].value = ''
        else :
            parameters[8].enabled = True

        return

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                desc = arcpy.Describe(inputFeatureClass)

                if (desc.shapeType != 'Point'):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be points feature class')
                else :
                    fieldList = arcpy.ListFields(inputFeatureClass)
                    b_gamma = False
                    b_line = False

                    for field in fieldList:
                        if field.name == 'recordedGamma':
                            b_gamma = True
                        elif field.name == 'lineNumber':
                            b_line = True

                    if b_gamma == False or b_line == False or b_alt == False:
                        parameters[0].setErrorMessage('Not all required input fields present lineNumber and recordedGamma')

        except Exception as e:
            return

        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input variables 
        inputFeatureClass = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputFeatureClass))

        outputWorkspace = parameters[8].valueAsText

        inputSurfaceMethod = parameters[1].value
        arcpy.AddMessage('inputSurfaceMethod: ' + inputSurfaceMethod)

        inputSurveyArea = parameters[2].valueAsText 
        arcpy.AddMessage('inputSurveyArea: ' + inputSurveyArea)

        if (inputSurfaceMethod == "Kriging") :
            inputLineSpacing = parameters[3].value
            arcpy.AddMessage('inputLineSpacing: ' + str(inputLineSpacing))

        inputRawContourInterval = parameters[4].value
        arcpy.AddMessage('Raw Gamma Contour Interval: ' + str(inputRawContourInterval))

        inputGradientContourInterval = parameters[5].value
        arcpy.AddMessage('Gradient Gamma Contour Interval: ' + str(inputGradientContourInterval))

        inputCellSize = parameters[6].value
        arcpy.AddMessage('inputCellSize: ' + str(inputCellSize))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.env.extent = inputSurveyArea
        arcpy.AddMessage('Scratch Workspace Folder: ' + arcpy.env.scratchFolder)
        arcpy.AddMessage('Set environment variables') 

        suffix = ''
        suffix_w_raw_contour = ''
        suffix_w_gradient_contour = ''
        suffix_w_both_contours = ''

        if (inputSurfaceMethod == "Kriging") :
            suffix = '_Kriging_LS_' + str(int(inputLineSpacing)) + '_CS_' + str(int(inputCellSize))
        elif (inputSurfaceMethod == "TIN") :
            suffix = '_TIN_CS_' + str(int(inputCellSize))

        suffix_w_raw_contour = suffix + '_RC_' + str(int(inputRawContourInterval))
        suffix_w_gradient_contour = suffix + '_GC_' + str(int(inputGradientContourInterval))
        suffix_w_both_contour = suffix + '_RC_' + str(int(inputRawContourInterval)) + '_GC_' + str(int(inputGradientContourInterval))

        # Create the logTable
        logTable = 'VisualizationLog' + suffix_w_both_contour
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.AddField_management(logTable, 'Process', 'TEXT')
        arcpy.AddField_management(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
        logCursor.insertRow(('Survey Points: {}'.format(inputFeatureClass), datetime.datetime.now()))
        logCursor.insertRow(('Survey area: {}'.format(inputSurveyArea), datetime.datetime.now()))
        logCursor.insertRow(('Surface Generation Method: ' + inputSurfaceMethod, datetime.datetime.now()))

        if (inputSurfaceMethod == "Kriging") :
            logCursor.insertRow(('Survey line spacing: {}'.format(str(inputLineSpacing)), datetime.datetime.now()))

        logCursor.insertRow(('Raw Gamma Contour interval: {}'.format(str(inputRawContourInterval)), datetime.datetime.now()))
        logCursor.insertRow(('Gradient Gamma Contour interval: {}'.format(str(inputGradientContourInterval)), datetime.datetime.now()))
        logCursor.insertRow(('Cell size for output rasters: {}'.format(str(inputCellSize)), datetime.datetime.now()))

        # Check out the spatial analyst extension 
        arcpy.CheckOutExtension('Spatial') 
        arcpy.AddMessage('Checked out Spatial Analyst extension')
        logCursor.insertRow(('Checked out Spatail Analyst extension', datetime.datetime.now()))

        # Create gammaReadings object 
        gammaReadings = GammaReadings() 

        gammaReadings.readSurveyLinesFromFeatureClass(inputFeatureClass)

        # Create a feature class from the gamma readings 
        gammaNumpyArray = gammaReadings.calculateSurveyLineGradients() 
        gradientGammaFeatureClass = os.path.join(outputWorkspace, 'gradientGamma')
        if not arcpy.Exists(gradientGammaFeatureClass):
            arcpy.da.NumPyArrayToFeatureClass(gammaNumpyArray, gradientGammaFeatureClass, ['easting', 'northing'], arcpy.Describe(inputFeatureClass).spatialReference) 
            arcpy.AddMessage('Created gradientGamma Feature Class') 
        else :
            arcpy.AddMessage('gradientGamma Feature Class already exists') 
 

        rawGammaSurface = None
        gradientGammaSurface = None

        # Interpolate a surface of raw gamma values
        if (inputSurfaceMethod == "Kriging") :
            # Set the geoprocessing mask 
            arcpy.env.mask = inputSurveyArea 
            arcpy.AddMessage('Set analysis mask')
            logCursor.insertRow(('Set analysis mask', datetime.datetime.now()))
            
            krigingModel = KrigingModelOrdinary('CIRCULAR')
            
            rawGammaSurface = 'rawGamma' + suffix
            rawGamma = Kriging(inputFeatureClass, 'recordedGamma', krigingModel, str(inputCellSize), str(inputLineSpacing * 1.5))
            rawGamma.save(rawGammaSurface)		
            arcpy.AddMessage('Interpolated raw gamma by Kriging') 
            logCursor.insertRow(('Interpolated raw gamma by Kriging', datetime.datetime.now()))

            gradientGammaSurface = 'gradientGamma' + suffix
            gradientGamma = Kriging(gradientGammaFeatureClass, 'gradientGamma', krigingModel, str(inputCellSize), str(inputLineSpacing * 1.5))
            gradientGamma.save(gradientGammaSurface)		
            arcpy.AddMessage('Interpolated gradient gamma by Kriging') 
            logCursor.insertRow(('Interpolated gradient gamma by Kriging', datetime.datetime.now()))

        elif (inputSurfaceMethod == "TIN") :
            arcpy.CheckOutExtension("3D")
            arcpy.AddMessage('Checked out 3D extension') 
            logCursor.insertRow(('Checked out 3D extension', datetime.datetime.now()))

            try:
                desc = arcpy.Describe(inputFeatureClass)
                inputSpatialReference = desc.spatialReference
                rawTIN = arcpy.env.scratchFolder + '\\rawTin'
                arcpy.AddMessage('Creating TIN dataset: ' + rawTIN)
                arcpy.CreateTin_3d(rawTIN, inputSpatialReference, inputFeatureClass + " recordedGamma masspoints", "DELAUNAY") 
                arcpy.AddMessage('TIN dataset created')
                logCursor.insertRow(('TIN dataset created', datetime.datetime.now()))

                rawGammaSurface = 'rawGamma' + suffix
                rawGammaSurfaceTemp = rawGammaSurface + '_temp' 
                arcpy.env.mask = inputSurveyArea 
                arcpy.TinRaster_3d(rawTIN, rawGammaSurfaceTemp, "FLOAT", "LINEAR", "CELLSIZE " + str(inputCellSize), "1")
                
                rGS = ExtractByMask(rawGammaSurfaceTemp, inputSurveyArea)
                rGS.save(rawGammaSurface)
                
                arcpy.Delete_management(rawGammaSurfaceTemp)
                arcpy.AddMessage('TIN surface raster created')

                arcpy.Delete_management(rawTIN)
                arcpy.AddMessage('TIN deleted:' + rawTIN)

                gradientTIN = arcpy.env.scratchFolder + '\\gradientTin'
                arcpy.AddMessage('Creating TIN dataset: ' + gradientTIN)
                arcpy.CreateTin_3d(gradientTIN, inputSpatialReference, "gradientGamma gradientGamma masspoints", "DELAUNAY") 
                arcpy.AddMessage('TIN dataset created')
                logCursor.insertRow(('TIN dataset created', datetime.datetime.now()))

                gradientGammaSurface = 'gradientGamma' + suffix
                gradientGammaSurfaceTemp = gradientGammaSurface + '_temp' 
                arcpy.env.mask = inputSurveyArea 
                arcpy.TinRaster_3d(gradientTIN,  gradientGammaSurfaceTemp, "FLOAT", "LINEAR", "CELLSIZE " + str(inputCellSize), "1")
                
                temp = ExtractByMask(gradientGammaSurfaceTemp, inputSurveyArea)
                temp.save(gradientGammaSurface)
                
                arcpy.Delete_management(gradientGammaSurfaceTemp)
                arcpy.AddMessage('TIN surface raster created')

                arcpy.Delete_management(gradientTIN)
                arcpy.AddMessage('TIN deleted:' + gradientTIN)

            except arcpy.ExecuteError as e:
                arcpy.AddError('Tool Error Creating TIN dataset: ' + e.message)
                logCursor.insertRow(('tool error creating TIN dataset: ' + e.message, datetime.datetime.now()))
                return
            except Exception as e:
                arcpy.AddError('Error Creating TIN dataset: ' + e.message)
                logCursor.insertRow(('error creating TIN dataset: ' + e.message, datetime.datetime.now()))
                return
            arcpy.CheckInExtension("3D")
        else :
            arcpy.AddMessage('Unknown interpolation method: ' + inputSurfaceMethod)
            logCursor.insertRow(('Unknown Interpolation Method', datetime.datetime.now()))
            return

        arcpy.env.workspace = outputWorkspace 

        # Create contours of the raw gamma values 
        if arcpy.Exists(rawGammaSurface):
            Contour(rawGammaSurface, 'rawGamma' + suffix_w_raw_contour, inputRawContourInterval) 
            arcpy.AddMessage('Created raw gamma contour lines')
            logCursor.insertRow(('Created raw gamma contour lines', datetime.datetime.now()))
        else :
            arcpy.AddMessage('Could not find surface : ' + rawGammaSurface + ' in workspace: ' + arcpy.env.workspace)

        # Create contours of the gradient gamma values 
        if arcpy.Exists(gradientGammaSurface):
            Contour(gradientGammaSurface, 'gradientGamma' + suffix_w_gradient_contour, inputGradientContourInterval, inputGradientContourInterval / 2.0) 
            arcpy.AddMessage('Created gradient gamma contour lines')
            logCursor.insertRow(('Created gradient gamma contour lines', datetime.datetime.now()))
        else :
            arcpy.AddMessage('Could not find surface : ' + gradientGammaSurface + ' in workspace: ' + arcpy.env.workspace)

        # Delete the logCursor object
        del logCursor
        arcpy.CheckInExtension('Spatial') 

class SurveyAreaTool(object): 
    def __init__(self): 
        """Define the tool (tool name is the name of the class)."""
        self.label = 'Generate Survey Boundary'
        self.description = '' 
        self.canRunInBackground = True

    def getParameterInfo(self): 
        """Define parameter definitions"""
        param0 = arcpy.Parameter(displayName = 'Input Survey Line or Point Feature Class', 
                                 name = 'inputSurveyFeatures', 
                                 datatype = 'DEFeatureClass', 
                                 parameterType = 'Required',
                                 direction = 'Input')

        param1 = arcpy.Parameter(displayName = 'Buffer Distance (meter)', 
                                 name = 'inputBufferDistance', 
                                 datatype = 'Double', 
                                 parameterType = 'Required', 
                                 direction = 'Input') 

        param2 =  arcpy.Parameter(displayName = 'Send Outputs to same geodatabase as input?', 
                                  name = 'useInputGeodatabase', 
                                  datatype = 'GPBoolean', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        param3 = arcpy.Parameter(displayName = 'Existing Output File Geodatabase',
                                  name = 'outputWorkspace', 
                                  datatype = 'Workspace', 
                                  parameterType = 'Required', 
                                  direction = 'Input') 

        # Default values 
        param1.value = 5.0
        param2.value = True
        return [param0, param1, param2, param3] 

    def isLicensed(self): 
        """Set whether tool is licensed to execute."""
        try: 
            if arcpy.SetProduct('arcinfo') == "NotLicensed": 
                raise Exception 
        except Exception: 
            return False
        return True

    def updateParameters(self, parameters): 
        """Modify the values and properties of parameters before internal 
        validation is performed.  This method is called whenever a parameter 
        has been changed."""

        if (parameters[2].value == True):
            parameters[3].enabled = False

            if (parameters[0].altered and len(parameters[0].valueAsText) > 0) : 
                workspaceparts = os.path.split(parameters[0].valueAsText)
                parameters[3].value = workspaceparts[0]
            else :
                parameters[3].value = ''
        else :
            parameters[3].enabled = True

    def updateMessages(self, parameters): 
        """Modify the messages created by internal validation for each tool 
        parameter.  This method is called after internal validation."""
        
        if parameters[1].value < 3.0 :
            parameters[1].setErrorMessage('Buffer distance must be 3 meters or greater')

        try:
            arcpy.env.workspace = parameters[3].valueAsText
            if len(arcpy.env.workspace) > 0 and arcpy.Exists('surveyArea'):
                parameters[0].setWarningMessage('Buffer already defined. Datasets may be overwritten')
        except Exception as e:
            return

        try:
            inputFeatureClass = parameters[0].valueAsText
            if len(inputFeatureClass) > 0: 
                desc = arcpy.Describe(inputFeatureClass)

                if ((desc.shapeType != 'Point') and (desc.shapeType != 'Polyline')):
                    parameters[0].setErrorMessage('Incorrect input shape type, should be point or line feature class')

        except Exception as e:
            return

        return

    def execute(self, parameters, messages): 
        """The source code of the tool."""
        # Set input variables 
        inputSurveyFeatures = parameters[0].valueAsText 
        arcpy.AddMessage('input Survey Features: {}'.format(inputSurveyFeatures))

        inputBufferDistance = parameters[1].value
        arcpy.AddMessage('inputBufferDistance: {}'.format(str(inputBufferDistance)))

        outputWorkspace = parameters[3].valueAsText

        arcpy.AddMessage('output Geodatabase: {}'.format(outputWorkspace))

        # Set arcpy environment variables 
        arcpy.env.workspace = outputWorkspace 
        arcpy.env.overwriteOutput = True
        arcpy.AddMessage('Set environment variables') 

        # Create the logTable
        now = datetime.datetime.now()
        logTable = 'SurveyAreaLog_{}'.format(now.strftime('%Y%m%d%H%M%S'))
        arcpy.CreateTable_management(outputWorkspace, logTable)
        arcpy.AddField_management(logTable, 'Process', 'TEXT')
        arcpy.AddField_management(logTable, 'Completed', 'TEXT')

        # Create an insert cursor to write to the logTable
        logCursor = arcpy.da.InsertCursor(logTable,('Process', 'Completed'))

        # Write the input parameters to the logTable
        logCursor.insertRow(('Survey Features: {}'.format(inputSurveyFeatures), datetime.datetime.now()))
        logCursor.insertRow(('Buffer Distance: {}'.format(inputBufferDistance), datetime.datetime.now()))

        # Make convex hull 
        arcpy.MinimumBoundingGeometry_management(inputSurveyFeatures, 'convexHull', 'CONVEX_HULL') 
        arcpy.AddMessage('Made convex hull') 
        logCursor.insertRow(('Made convex hull', datetime.datetime.now()))

        # Create survey area by buffering convex hull 
        arcpy.Buffer_analysis('convexHull', 'surveyArea', inputBufferDistance, 'FULL', '', 'ALL') 
        arcpy.AddMessage('Made survey area by buffering convex hull') 
        logCursor.insertRow(('Made survey area by buffering convex hull', datetime.datetime.now()))

        arcpy.AddMessage('Making survey area 2') 

        arcpy.Buffer_analysis(inputSurveyFeatures, 'surveyArea_temp1', inputBufferDistance, 'FULL', 'ROUND', 'ALL')

        arcpy.AddMessage('bufferred individual input features') 
        logCursor.insertRow(('bufferred individual input features', datetime.datetime.now()))

        try:
        
            arcpy.EliminatePolygonPart_management('surveyArea_temp1', 'surveyArea_temp2', 'PERCENT', '', 99.0, 'CONTAINED_ONLY')

            arcpy.AddMessage('Eliminated internal voids in search area') 
            logCursor.insertRow(('Eliminated internal voids in search area', datetime.datetime.now()))
        except Exception:
            arcpy.AddMessage('Unable to eliminate internal voids in search area') 
            logCursor.insertRow(('Unable to eliminate internal voids in search area', datetime.datetime.now()))

        tempfeature = None

        try:
            if arcpy.Exists('surveyArea_temp2'):
                tempfeature = 'surveyArea_temp2'
            else:
                tempfeature = 'surveyArea_temp1'

            arcpy.SimplifyPolygon_cartography(tempfeature, 'surveyArea2', "POINT_REMOVE", (inputBufferDistance / 4.0), 0, 'RESOLVE_ERRORS', 'NO_KEEP')

            arcpy.AddMessage('Simplified bufferred survey area') 
            logCursor.insertRow(('Simplified bufferred survey area', datetime.datetime.now()))
        except Exception:
            arcpy.AddMessage('Unable to simplify survey area') 
            logCursor.insertRow(('Unable to simplify survey area', datetime.datetime.now()))

            arcpy.Rename_management(tempfeature, 'surveyArea2')

        if arcpy.Exists('surveyArea2' + '_Pnt'):
            arcpy.Delete_management('surveyArea2' + '_Pnt')

        if arcpy.Exists('surveyArea_temp1'):
            arcpy.Delete_management('surveyArea_temp1')
        
        if arcpy.Exists('surveyArea_temp2'):
            arcpy.Delete_management('surveyArea_temp2')

        logCursor.insertRow(('Made survey area 2', datetime.datetime.now()))

        # Delete the logCursor object
        del logCursor